#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Iterable, Optional, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.spider_max import spider_max


pylog = logging.getLogger(__name__)


class SPIDErMax(AACMetric):
    """SPIDEr-max class.

    - Paper: https://hal.archives-ouvertes.fr/hal-03810396/file/Labbe_DCASE2022.pdf

    For more information, see :func:`~aac_metrics.functional.spider.spider`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 5.5

    def __init__(
        self,
        return_all_scores: bool = True,
        return_all_cands_scores: bool = False,
        # CIDEr args
        n: int = 4,
        sigma: float = 6.0,
        # SPICE args
        cache_path: str = "$HOME/.cache",
        java_path: str = "java",
        tmp_path: str = "/tmp",
        n_threads: Optional[int] = None,
        java_max_memory: str = "8G",
        timeout: Union[None, int, Iterable[int]] = None,
        verbose: int = 0,
    ) -> None:
        super().__init__()
        self._return_all_scores = return_all_scores
        self._return_all_cands_scores = return_all_cands_scores
        self._n = n
        self._sigma = sigma
        self._cache_path = cache_path
        self._java_path = java_path
        self._tmp_path = tmp_path
        self._n_threads = n_threads
        self._java_max_memory = java_max_memory
        self._timeout = timeout
        self._verbose = verbose

        self._mult_candidates = []
        self._mult_references = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return spider_max(
            self._mult_candidates,
            self._mult_references,
            self._return_all_scores,
            self._return_all_cands_scores,
            n=self._n,
            sigma=self._sigma,
            cache_path=self._cache_path,
            java_path=self._java_path,
            tmp_path=self._tmp_path,
            n_threads=self._n_threads,
            java_max_memory=self._java_max_memory,
            timeout=self._timeout,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        return (
            f"n={self._n}, sigma={self._sigma}, java_max_memory={self._java_max_memory}"
        )

    def get_output_names(self) -> tuple[str, ...]:
        return ("spider_max",)

    def reset(self) -> None:
        self._mult_candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        mult_candidates: list[list[str]],
        mult_references: list[list[str]],
    ) -> None:
        self._mult_candidates += mult_candidates
        self._mult_references += mult_references



# =================================================================
# This code was pulled from https://github.com/tylin/coco-caption
# Image-specific names and comments have been changed to be audio-specific
# =================================================================

from __future__ import division
import os
import sys
import subprocess
import threading
import json
import numpy as np
import ast
import tempfile

# Assumes spice.jar is in the same directory as spice.py.  Change as needed.
SPICE_JAR = 'spice-1.0.jar'
TEMP_DIR = 'tmp'
CACHE_DIR = 'cache'

class Spice:
    """
    Main Class to compute the SPICE metric 
    """

    def float_convert(self, obj):
        try:
          return float(obj)
        except:
          return np.nan

    def compute_score(self, gts, res):
        assert(sorted(gts.keys()) == sorted(res.keys()))
        audioIds = sorted(gts.keys())
        
        # Prepare temp input file for the SPICE scorer
        input_data = []
        for id in audioIds:
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert(type(hypo) is list)
            assert(len(hypo) == 1)
            assert(type(ref) is list)
            assert(len(ref) >= 1)

            input_data.append({
              "image_id" : id,      # This has to be "image_id" for SPICE evaluation to work
              "test" : hypo[0],
              "refs" : ref
            })

        cwd = os.path.dirname(os.path.abspath(__file__))
        temp_dir=os.path.join(cwd, TEMP_DIR)
        if not os.path.exists(temp_dir):
          os.makedirs(temp_dir)
        in_file = tempfile.NamedTemporaryFile(delete=False, dir=temp_dir, mode='w')
        json.dump(input_data, in_file, indent=2)
        in_file.close()

        # Start job
        out_file = tempfile.NamedTemporaryFile(delete=False, dir=temp_dir)
        out_file.close()
        cache_dir=os.path.join(cwd, CACHE_DIR)
        if not os.path.exists(cache_dir):
          os.makedirs(cache_dir)
        spice_cmd = ['java', '-jar', '-Xmx8G', SPICE_JAR, in_file.name,
          '-cache', cache_dir,
          '-out', out_file.name,
          '-subset',
          '-silent'
        ]
        subprocess.check_call(spice_cmd, 
            cwd=os.path.dirname(os.path.abspath(__file__)))

        # Read and process results
        with open(out_file.name) as data_file:    
          results = json.load(data_file)
        os.remove(in_file.name)
        os.remove(out_file.name)

        audioId_to_scores = {}
        spice_scores = []
        for item in results:
          audioId_to_scores[item['image_id']] = item['scores']
          spice_scores.append(self.float_convert(item['scores']['All']['f']))
        average_score = np.mean(np.array(spice_scores))
        scores = []
        for audio_id in audioIds:
          # Convert none to NaN before saving scores over subcategories
          score_set = {}
          for category,score_tuple in audioId_to_scores[audio_id].items():
            score_set[category] = {k: self.float_convert(v) for k, v in score_tuple.items()}
          scores.append(score_set)
        return average_score, scores

    def method(self):
        return "SPICE"





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from unittest import TestCase

from aac_metrics.classes.bleu import BLEU
from aac_metrics.utils.imports import _TORCHMETRICS_AVAILABLE

if _TORCHMETRICS_AVAILABLE:
    from torchmetrics.text.bleu import BLEUScore


class TestBleu(TestCase):
    # Tests methods
    def test_bleu(self) -> None:
        if not _TORCHMETRICS_AVAILABLE:
            return None

        cands = ["a man is speaking", "birds chirping"]
        mrefs = [
            [
                "man speaks",
                "man is speaking",
                "a man speaks",
                "man talks",
                "someone is talking",
            ],
            ["a bird is chirping"] * 5,
        ]

        n = 2

        bleu_v1 = BLEU(n=n, return_all_scores=False)
        score_v1 = bleu_v1(cands, mrefs)

        bleu_v2 = BLEUScore(n_gram=n, smooth=False)
        score_v2 = bleu_v2(cands, mrefs)

        self.assertAlmostEqual(score_v1.item(), score_v2.item())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
BASED ON https://github.com/blmoistawinde/fense/
"""

import logging

from typing import Union

import numpy as np
import torch

from sentence_transformers import SentenceTransformer
from torch import Tensor


pylog = logging.getLogger(__name__)


def sbert(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
    device: Union[str, torch.device, None] = "auto",
    batch_size: int = 32,
    reset_state: bool = True,
    verbose: int = 0,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Tensor]]]:
    """Cosine-similarity of the Sentence-BERT embeddings.

    - Paper: https://arxiv.org/abs/1908.10084
    - Original implementation: https://github.com/blmoistawinde/fense

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param sbert_model: The sentence BERT model used to extract sentence embeddings for cosine-similarity. defaults to "paraphrase-TinyBERT-L6-v2".
    :param device: The PyTorch device used to run FENSE models. If "auto", it will use cuda if available. defaults to "cpu".
    :param batch_size: The batch size of the sBERT models. defaults to 32.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    # Init models
    sbert_model = _load_sbert(sbert_model, device, reset_state)

    # Encode sents
    rng_ids = [0]
    for refs in mult_references:
        rng_ids.append(rng_ids[-1] + len(refs))
    flat_references = [ref for refs in mult_references for ref in refs]

    cands_embs = _encode_sents_sbert(sbert_model, candidates, batch_size, verbose)
    mrefs_embs = _encode_sents_sbert(sbert_model, flat_references, batch_size, verbose)

    # Compute sBERT similarities
    sbert_cos_sims = [
        (cands_embs[i] @ mrefs_embs[rng_ids[i] : rng_ids[i + 1]].T).mean().item()
        for i in range(len(cands_embs))
    ]
    sbert_cos_sims = np.array(sbert_cos_sims)

    # Aggregate and return
    sbert_cos_sim = sbert_cos_sims.mean()

    sbert_cos_sim = torch.as_tensor(sbert_cos_sim)
    sbert_cos_sims = torch.from_numpy(sbert_cos_sims)

    if return_all_scores:
        sents_scores = {
            "sbert.sim": sbert_cos_sims,
        }
        corpus_scores = {
            "sbert.sim": sbert_cos_sim,
        }

        return corpus_scores, sents_scores
    else:
        return sbert_cos_sim


def _load_sbert(
    sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
    device: Union[str, torch.device, None] = "auto",
    reset_state: bool = True,
) -> SentenceTransformer:
    state = torch.random.get_rng_state()

    if device == "auto":
        device = "cuda" if torch.cuda.is_available() else "cpu"
    if isinstance(device, str):
        device = torch.device(device)

    if isinstance(sbert_model, str):
        sbert_model = SentenceTransformer(sbert_model, device=device)  # type: ignore
    sbert_model.to(device=device)

    sbert_model = sbert_model.eval()
    for p in sbert_model.parameters():
        p.requires_grad_(False)

    if reset_state:
        torch.random.set_rng_state(state)
    return sbert_model


@torch.no_grad()
def _encode_sents_sbert(
    sbert_model: SentenceTransformer,
    sents: list[str],
    batch_size: int = 32,
    verbose: int = 0,
) -> Tensor:
    return sbert_model.encode(
        sents,
        convert_to_tensor=True,
        normalize_embeddings=True,
        batch_size=batch_size,
        show_progress_bar=verbose >= 2,
    )  # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import logging
import sys

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Iterable, Union

import yaml

from aac_metrics.functional.evaluate import evaluate, METRICS_SETS
from aac_metrics.utils.checks import check_metric_inputs, check_java_path


pylog = logging.getLogger(__name__)


def load_csv_file(
    fpath: Union[str, Path],
    cands_columns: Union[str, Iterable[str]] = ("caption_predicted",),
    mrefs_columns: Union[str, Iterable[str]] = (
        "caption_1",
        "caption_2",
        "caption_3",
        "caption_4",
        "caption_5",
    ),
    load_mult_cands: bool = False,
    strict: bool = True,
) -> tuple[list, list[list[str]]]:
    """Load candidates and mult_references from a CSV file.

    :param fpath: The filepath to the CSV file.
    :param cands_columns: The columns of the candidates. defaults to ("captions_predicted",).
    :param mrefs_columns: The columns of the multiple references. defaults to ("caption_1", "caption_2", "caption_3", "caption_4", "caption_5").
    :param load_mult_cands: If True, load multiple candidates from file. defaults to False.
    :returns: A tuple of (candidates, mult_references) loaded from file.
    """
    if isinstance(cands_columns, str):
        cands_columns = [cands_columns]
    else:
        cands_columns = list(cands_columns)

    if isinstance(mrefs_columns, str):
        mrefs_columns = [mrefs_columns]
    else:
        mrefs_columns = list(mrefs_columns)

    with open(fpath, "r") as file:
        reader = csv.DictReader(file)
        fieldnames = reader.fieldnames
        data = list(reader)

    if fieldnames is None:
        raise ValueError(f"Cannot read fieldnames in CSV file {fpath=}.")

    file_cands_columns = [column for column in cands_columns if column in fieldnames]
    file_mrefs_columns = [column for column in mrefs_columns if column in fieldnames]

    if strict:
        if len(file_cands_columns) != len(cands_columns):
            raise ValueError(
                f"Cannot find all candidates columns {cands_columns=} in file '{fpath}'."
            )
        if len(file_mrefs_columns) != len(mrefs_columns):
            raise ValueError(
                f"Cannot find all references columns {mrefs_columns=} in file '{fpath}'."
            )

    if (load_mult_cands and len(file_cands_columns) <= 0) or (
        not load_mult_cands and len(file_cands_columns) != 1
    ):
        raise ValueError(
            f"Cannot find candidate column in file. ({cands_columns=} not found in {fieldnames=})"
        )
    if len(file_mrefs_columns) <= 0:
        raise ValueError(
            f"Cannot find references columns in file. ({mrefs_columns=} not found in {fieldnames=})"
        )

    if load_mult_cands:
        mult_candidates = _load_columns(data, file_cands_columns)
        mult_references = _load_columns(data, file_mrefs_columns)
        return mult_candidates, mult_references
    else:
        file_cand_column = file_cands_columns[0]
        candidates = [data_i[file_cand_column] for data_i in data]
        mult_references = _load_columns(data, file_mrefs_columns)
        return candidates, mult_references


def _load_columns(data: list[dict[str, str]], columns: list[str]) -> list[list[str]]:
    mult_sentences = []
    for data_i in data:
        raw_sents = [data_i[column] for column in columns]
        sents = []
        for raw_sent in raw_sents:
            # Refs columns can be list[str]
            if "[" in raw_sent and "]" in raw_sent:
                try:
                    sent = eval(raw_sent)
                    assert isinstance(sent, list) and all(
                        isinstance(sent_i, str) for sent_i in sent
                    )
                    sents += sent
                except (SyntaxError, NameError):
                    sents.append(raw_sent)
            else:
                sents.append(raw_sent)

        mult_sentences.append(sents)
    return mult_sentences


def _get_main_evaluate_args() -> Namespace:
    parser = ArgumentParser(description="Evaluate an output file.")

    parser.add_argument(
        "--input_file",
        "-i",
        type=str,
        default="",
        help="The input file path containing the candidates and references.",
        required=True,
    )
    parser.add_argument(
        "--cand_columns",
        "-cc",
        type=str,
        nargs="+",
        default=("caption_predicted", "preds", "cands"),
        help="The column names of the candidates in the CSV file. defaults to ('caption_predicted', 'preds', 'cands').",
    )
    parser.add_argument(
        "--mrefs_columns",
        "-rc",
        type=str,
        nargs="+",
        default=(
            "caption_1",
            "caption_2",
            "caption_3",
            "caption_4",
            "caption_5",
            "captions",
        ),
        help="The column names of the candidates in the CSV file. defaults to ('caption_1', 'caption_2', 'caption_3', 'caption_4', 'caption_5', 'captions').",
    )
    parser.add_argument(
        "--metrics_set_name",
        type=str,
        default="default",
        choices=tuple(METRICS_SETS.keys()),
        help=f"The metrics set to compute. Can be one of {tuple(METRICS_SETS.keys())}. defaults to 'default'.",
    )
    parser.add_argument(
        "--cache_path",
        type=str,
        default="$HOME/.cache",
        help="Cache directory path. defaults to '$HOME/.cache'.",
    )
    parser.add_argument(
        "--java_path",
        type=str,
        default="java",
        help="Java executable path. defaults to 'java'.",
    )
    parser.add_argument(
        "--tmp_path",
        type=str,
        default="/tmp",
        help="Temporary directory path. defaults to '/tmp'.",
    )
    parser.add_argument("--verbose", type=int, default=0, help="Verbose level.")

    args = parser.parse_args()
    return args


def _main_evaluate() -> None:
    format_ = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format_))
    pkg_logger = logging.getLogger("aac_metrics")
    pkg_logger.addHandler(handler)

    args = _get_main_evaluate_args()

    if not check_java_path(args.java_path):
        raise RuntimeError(f"Invalid Java executable. ({args.java_path})")

    level = logging.INFO if args.verbose <= 1 else logging.DEBUG
    pkg_logger.setLevel(level)

    if args.verbose >= 1:
        pylog.info(f"Load file {args.input_file}...")

    candidates, mult_references = load_csv_file(
        args.input_file, args.cand_columns, args.mrefs_columns
    )
    check_metric_inputs(candidates, mult_references)

    refs_lens = list(map(len, mult_references))
    if args.verbose >= 1:
        pylog.info(
            f"Found {len(candidates)} candidates, {len(mult_references)} references and [{min(refs_lens)}, {max(refs_lens)}] references per candidate."
        )

    corpus_scores, _sents_scores = evaluate(
        candidates,
        mult_references,
        True,
        args.metrics_set_name,
        args.cache_path,
        args.java_path,
        args.tmp_path,
        args.verbose,
    )

    corpus_scores = {k: v.item() for k, v in corpus_scores.items()}
    pylog.info(f"Global scores:\n{yaml.dump(corpus_scores, sort_keys=False)}")


if __name__ == "__main__":
    _main_evaluate()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pickle
import unittest

from unittest import TestCase

from aac_metrics.classes.evaluate import _instantiate_metrics_classes


class TestCompare(TestCase):
    # Tests methods
    def test_pickle_dump(self) -> None:
        metrics = _instantiate_metrics_classes("all")

        for metric in metrics:
            try:
                pickle.dumps(metric)
            except pickle.PicklingError:
                self.assert_(False, f"Cannot pickle {metric.__class__.__name__}.")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import math
import os
import os.path as osp
import shutil
import subprocess
import tempfile
import time

from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile
from typing import Any, Iterable, Optional, Union

import numpy as np
import torch

from torch import Tensor

from aac_metrics.utils.checks import check_java_path


pylog = logging.getLogger(__name__)


DNAME_SPICE_CACHE = osp.join("aac-metrics", "spice", "cache")
FNAME_SPICE_JAR = osp.join("aac-metrics", "spice", "spice-1.0.jar")


def spice(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    n_threads: Optional[int] = None,
    java_max_memory: str = "8G",
    timeout: Union[None, int, Iterable[int]] = None,
    separate_cache_dir: bool = True,
    verbose: int = 0,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """Semantic Propositional Image Caption Evaluation function.

    - Paper: https://arxiv.org/pdf/1607.08822.pdf

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param java_path: The path to the java executable. defaults to "java".
    :param tmp_path: Temporary directory path. defaults to "/tmp".
    :param n_threads: Number of threads used to compute SPICE.
        None value will use the default value of the java program.
        defaults to None.
    :param java_max_memory: The maximal java memory used. defaults to "8G".
    :param timeout: The number of seconds before killing the java subprogram.
        If a list is given, it will restart the program if the i-th timeout is reached.
        If None, no timeout will be used.
        defaults to None.
    :param separate_cache_dir: If True, the SPICE cache files will be stored into in a new temporary directory.
        This removes potential freezes when multiple instances of SPICE are running in the same cache dir.
        defaults to True.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """

    cache_path = osp.expandvars(cache_path)
    java_path = osp.expandvars(java_path)
    tmp_path = osp.expandvars(tmp_path)

    spice_fpath = osp.join(cache_path, FNAME_SPICE_JAR)

    if __debug__:
        if not osp.isfile(spice_fpath):
            raise FileNotFoundError(
                f"Cannot find JAR file '{spice_fpath}' for SPICE metric. Maybe run 'aac-metrics-download' or specify another 'cache_path' directory."
            )
        if not check_java_path(java_path):
            raise RuntimeError(
                f"Invalid Java executable to compute SPICE score. ({java_path})"
            )

    if len(candidates) != len(mult_references):
        raise ValueError(
            f"Invalid number of candidates and references. (found {len(candidates)=} != {len(mult_references)=})"
        )

    if separate_cache_dir:
        spice_cache = tempfile.mkdtemp(dir=tmp_path)
    else:
        spice_cache = osp.join(cache_path, DNAME_SPICE_CACHE)
    del cache_path

    if verbose >= 2:
        pylog.debug(f"Use cache directory {spice_cache}.")
        pylog.debug(f"Computing SPICE with JAR file {spice_fpath}...")

    input_data = [
        {
            "image_id": i,
            "test": cand,
            "refs": refs,
        }
        for i, (cand, refs) in enumerate(zip(candidates, mult_references))
    ]

    in_file = NamedTemporaryFile(
        mode="w", delete=False, dir=tmp_path, prefix="spice_inputs_", suffix=".json"
    )
    json.dump(input_data, in_file, indent=2)
    in_file.close()

    # Sometimes the java program can freeze, so timeout has been added to avoid using job time.
    if timeout is None or isinstance(timeout, (int, float)):
        timeout_lst = [timeout]
    else:
        timeout_lst = list(timeout)

    out_file = NamedTemporaryFile(
        mode="w", delete=False, dir=tmp_path, prefix="spice_outputs_", suffix=".json"
    )
    out_file.close()

    for i, timeout_i in enumerate(timeout_lst):
        if verbose >= 3:
            stdout = None
            stderr = None
        else:
            stdout = NamedTemporaryFile(
                mode="w",
                delete=True,
                dir=tmp_path,
                prefix="spice_stdout_",
                suffix=".txt",
            )
            stderr = NamedTemporaryFile(
                mode="w",
                delete=True,
                dir=tmp_path,
                prefix="spice_stderr_",
                suffix=".txt",
            )

        spice_cmd = [
            java_path,
            "-jar",
            f"-Xmx{java_max_memory}",
            spice_fpath,
            in_file.name,
            "-cache",
            spice_cache,
            "-out",
            out_file.name,
            "-subset",
        ]
        if n_threads is not None:
            spice_cmd += ["-threads", str(n_threads)]

        if verbose >= 2:
            pylog.debug(f"Run SPICE java code with: {' '.join(spice_cmd)}")

        try:
            subprocess.check_call(
                spice_cmd,
                stdout=stdout,
                stderr=stderr,
                timeout=timeout_i,
            )
            if stdout is not None:
                stdout.close()
            if stderr is not None:
                stderr.close()
            break

        except subprocess.TimeoutExpired as err:
            pylog.warning(
                f"Timeout SPICE java program with {timeout_i=}s (nb timeouts done={i+1}/{len(timeout_lst)})."
            )

            if i < len(timeout_lst) - 1:
                # Clear out files
                open(out_file.name, "w").close()
                if stdout is not None:
                    stdout.close()
                if stderr is not None:
                    stderr.close()
                time.sleep(1.0)
            else:
                raise err

        except (CalledProcessError, PermissionError) as err:
            pylog.error("Invalid SPICE call.")
            pylog.error(f"Full command: '{' '.join(spice_cmd)}'")
            if (
                stdout is not None
                and stderr is not None
                and osp.isfile(stdout.name)
                and osp.isfile(stderr.name)
            ):
                stdout_crashlog = stdout.name.replace(
                    "spice_stdout", "CRASH_spice_stdout"
                )
                stderr_crashlog = stderr.name.replace(
                    "spice_stderr", "CRASH_spice_stderr"
                )
                shutil.copy(stdout.name, stdout_crashlog)
                shutil.copy(stderr.name, stderr_crashlog)
                pylog.error(
                    f"For more information, see temp files '{stdout_crashlog}' and '{stderr_crashlog}'."
                )
            else:
                pylog.info(
                    f"Note: No temp file recorded. (found {stdout=} and {stderr=})"
                )
            raise err

    if verbose >= 2:
        pylog.debug("SPICE java code finished.")

    # Read and process results
    with open(out_file.name, "r") as data_file:
        results = json.load(data_file)
    os.remove(in_file.name)
    os.remove(out_file.name)

    if separate_cache_dir:
        shutil.rmtree(spice_cache)

    imgId_to_scores = {}
    spice_scores = []
    for item in results:
        imgId_to_scores[item["image_id"]] = item["scores"]
        spice_scores.append(__float_convert(item["scores"]["All"]["f"]))

    spice_scores = np.array(spice_scores)
    # Note: use numpy to compute mean because np.mean and torch.mean can give very small differences
    spice_score = spice_scores.mean()

    dtype = torch.float64
    spice_scores = torch.from_numpy(spice_scores)
    spice_score = torch.as_tensor(spice_score, dtype=dtype)

    if return_all_scores:
        spice_outs_corpus = {
            "spice": spice_score,
        }
        spice_outs_sents = {
            "spice": spice_scores,
        }
        return spice_outs_corpus, spice_outs_sents
    else:
        return spice_score


def __float_convert(obj: Any) -> float:
    try:
        return float(obj)
    except (ValueError, TypeError):
        return math.nan



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from functools import cache
from importlib.util import find_spec


@cache
def _package_is_available(package_name: str) -> bool:
    try:
        return find_spec(package_name) is not None
    except AttributeError:
        # Python 3.6
        return False
    except (ImportError, ModuleNotFoundError):
        # Python 3.7+
        return False


_TORCHMETRICS_AVAILABLE: bool = _package_is_available("torchmetrics")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Union

import torch

from sentence_transformers import SentenceTransformer
from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.sbert_sim import sbert_sim, _load_sbert


pylog = logging.getLogger(__name__)


class SBERTSim(AACMetric):
    """Cosine-similarity of the Sentence-BERT embeddings.

    - Paper: https://arxiv.org/abs/1908.10084
    - Original implementation: https://github.com/blmoistawinde/fense

    For more information, see :func:`~aac_metrics.functional.sbert.sbert`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = -1.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
        device: Union[str, torch.device, None] = "auto",
        batch_size: int = 32,
        reset_state: bool = True,
        verbose: int = 0,
    ) -> None:
        sbert_model = _load_sbert(sbert_model, device, reset_state)

        super().__init__()
        self._return_all_scores = return_all_scores
        self._sbert_model = sbert_model
        self._device = device
        self._batch_size = batch_size
        self._reset_state = reset_state
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return sbert_sim(
            self._candidates,
            self._mult_references,
            self._return_all_scores,
            self._sbert_model,
            self._device,
            self._batch_size,
            self._reset_state,
            self._verbose,
        )

    def extra_repr(self) -> str:
        return f"device={self._device}, batch_size={self._batch_size}"

    def get_output_names(self) -> tuple[str, ...]:
        return ("sbert.sim",)

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._candidates += candidates
        self._mult_references += mult_references



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import sys

from typing import Dict

import torch
import yaml

import aac_metrics


def get_install_info() -> Dict[str, str]:
    """Return a dictionary containing the version python, the os name, the architecture name and the versions of the following packages: aac_datasets, torch, torchaudio."""
    return {
        "aac_metrics": aac_metrics.__version__,
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "os": platform.system(),
        "architecture": platform.architecture()[0],
        "torch": str(torch.__version__),
    }


def print_install_info() -> None:
    """Print packages versions and architecture info."""
    install_info = get_install_info()
    print(yaml.dump(install_info, sort_keys=False))


if __name__ == "__main__":
    print_install_info()



# =================================================================
# This code was pulled from https://github.com/tylin/coco-caption
# and refactored for Python 3 and to also evaluate the SPIDEr metric.
# Image-specific names and comments have also been changed to be audio-specific
# =================================================================

__author__ = 'tylin'
from .tokenizer.ptbtokenizer import PTBTokenizer
from .bleu.bleu import Bleu
from .meteor.meteor import Meteor
from .rouge.rouge import Rouge
from .cider.cider import Cider
from .spice.spice import Spice
import numpy as np

class COCOEvalCap:
    def __init__(self, coco, cocoRes):
        self.evalAudios = []
        self.eval = {}
        self.audioToEval = {}
        self.coco = coco
        self.cocoRes = cocoRes
        self.params = {'audio_id': coco.getAudioIds()}

    def evaluate(self, verbose=True):
        audioIds = self.params['audio_id']
        # audioIds = self.coco.getAudioIds()
        gts = {}
        res = {}
        for audioId in audioIds:
            gts[audioId] = self.coco.audioToAnns[audioId]
            res[audioId] = self.cocoRes.audioToAnns[audioId]

        # =================================================
        # Set up scorers
        # =================================================
        if verbose:
            print('tokenization...')
        tokenizer = PTBTokenizer()
        gts = tokenizer.tokenize(gts)
        res = tokenizer.tokenize(res)

        # =================================================
        # Set up scorers
        # =================================================
        if verbose:
            print('setting up scorers...')
        scorers = [
            (Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"]),
            (Meteor(), "METEOR"),
            (Rouge(), "ROUGE_L"),
            (Cider(), "CIDEr"),
            (Spice(), "SPICE")
        ]

        # =================================================
        # Compute scores
        # =================================================
        for scorer, method in scorers:
            if verbose:
                print('computing %s score...'%(scorer.method()))
            score, scores = scorer.compute_score(gts, res)
            if type(method) == list:
                for sc, scs, m in zip(score, scores, method):
                    self.setEval(sc, m)
                    self.setAudioToEvalAudios(scs, gts.keys(), m)
                    if verbose:
                        print("%s: %0.3f"%(m, sc))
            else:
                self.setEval(score, method)
                self.setAudioToEvalAudios(scores, gts.keys(), method)
                if verbose:
                    print("%s: %0.3f"%(method, score))

        # Compute SPIDEr metric (average of CIDEr and SPICE)
        if verbose:
            print('computing %s score...' % ('SPIDEr'))
        score = (self.eval['CIDEr'] + self.eval['SPICE']) / 2.
        scores = list((
                np.array([audio['CIDEr'] for audio in self.audioToEval.values()]) +
                np.array([audio['SPICE']['All']['f'] for audio in self.audioToEval.values()])
        ) / 2)
        self.setEval(score, 'SPIDEr')
        self.setAudioToEvalAudios(scores, gts.keys(), 'SPIDEr')
        if verbose:
            print("%s: %0.3f" % ('SPIDEr', score))

        self.setEvalAudios()

    def setEval(self, score, method):
        self.eval[method] = score

    def setAudioToEvalAudios(self, scores, audioIds, method):
        for audioId, score in zip(audioIds, scores):
            if not audioId in self.audioToEval:
                self.audioToEval[audioId] = {}
                self.audioToEval[audioId]["audio_id"] = audioId
            self.audioToEval[audioId][method] = score

    def setEvalAudios(self):
        self.evalAudios = [eval for audioId, eval in self.audioToEval.items()]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import subprocess

from functools import cache
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Union


pylog = logging.getLogger(__name__)

MIN_JAVA_MAJOR_VERSION = 8
MAX_JAVA_MAJOR_VERSION = 11


def check_metric_inputs(
    candidates: Any,
    mult_references: Any,
) -> None:
    """Raises ValueError if candidates and mult_references does not have a valid type and size."""
    if not is_mono_sents(candidates):
        raise ValueError("Invalid candidates type. (expected list[str])")

    if not is_mult_sents(mult_references):
        raise ValueError("Invalid mult_references type. (expected list[list[str]])")

    same_len = len(candidates) == len(mult_references)
    if not same_len:
        raise ValueError(
            f"Invalid number of candidates ({len(candidates)}) with the number of references ({len(mult_references)})."
        )

    at_least_1_ref_per_cand = all(len(refs) > 0 for refs in mult_references)
    if not at_least_1_ref_per_cand:
        raise ValueError(
            "Invalid number of references per candidate. (found at least 1 empty list of references)"
        )


def check_java_path(java_path: Union[str, Path]) -> bool:
    """Returns True if the java path is valid."""
    if not isinstance(java_path, (str, Path)):
        return False

    output = ""
    try:
        output = subprocess.check_output(
            [str(java_path), "--version"],
        )
        output = output.decode().strip()
        version = output.split("\n")[0]
        major_version = int(version.split(" ")[1].split(".")[0])

    except (
        CalledProcessError,
        PermissionError,
        FileNotFoundError,
    ) as err:
        pylog.error(f"Invalid java path. (from {java_path=} and found error={err})")
        return False

    except (
        IndexError,
        ValueError,
    ) as err:
        pylog.error(f"Invalid java version. (found {output=} and {err=})")
        return False

    if not (MIN_JAVA_MAJOR_VERSION <= major_version <= MAX_JAVA_MAJOR_VERSION):
        pylog.error(
            f"Using Java version {version} is not officially supported by aac-metrics package and could not work for METEOR and SPICE metrics."
            f"(found {major_version=} but expected in [{MIN_JAVA_MAJOR_VERSION}, {MAX_JAVA_MAJOR_VERSION}])"
        )
        return False

    return True


def is_mono_sents(sents: Any) -> bool:
    """Returns True if input is list[str]."""
    return isinstance(sents, list) and all(isinstance(sent, str) for sent in sents)


def is_mult_sents(mult_sents: Any) -> bool:
    """Returns True if input is list[list[str]]."""
    return (
        isinstance(mult_sents, list)
        and all(isinstance(sents, list) for sents in mult_sents)
        and all(isinstance(sent, str) for sents in mult_sents for sent in sents)
    )


@cache
def _warn_once(msg: str) -> None:
    pylog.warning(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.rouge_l import (
    _rouge_l_compute,
    _rouge_l_update,
)


class ROUGEL(AACMetric):
    """Recall-Oriented Understudy for Gisting Evaluation class.

    - Paper: https://aclanthology.org/W04-1013.pdf

    For more information, see :func:`~aac_metrics.functional.rouge_l.rouge_l`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        beta: float = 1.2,
        tokenizer: Callable[[str], list[str]] = str.split,
    ) -> None:
        super().__init__()
        self._return_all_scores = return_all_scores
        self._beta = beta
        self._tokenizer = tokenizer

        self._rouge_l_scores = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return _rouge_l_compute(
            self._rouge_l_scores,
            self._return_all_scores,
        )

    def extra_repr(self) -> str:
        return f"beta={self._beta}"

    def get_output_names(self) -> tuple[str, ...]:
        return ("rouge_l",)

    def reset(self) -> None:
        self._rouge_l_scores = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._rouge_l_scores = _rouge_l_update(
            candidates,
            mult_references,
            self._beta,
            self._tokenizer,
            self._rouge_l_scores,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Optional

from aac_metrics.utils.imports import _TORCHMETRICS_AVAILABLE


if _TORCHMETRICS_AVAILABLE:
    from torchmetrics import Metric as __BaseMetric  # type: ignore

    class AACMetric(__BaseMetric):  # type: ignore
        # The theorical minimal value of the main global score of the metric.
        min_value: Optional[float] = None
        # The theorical maximal value of the main global score of the metric.
        max_value: Optional[float] = None

else:
    from torch import nn

    class AACMetric(nn.Module):
        """Base Metric module used when torchmetrics is not installed."""

        # Global values
        full_state_update: Optional[bool] = False
        higher_is_better: Optional[bool] = None
        is_differentiable: Optional[bool] = False

        # The theorical minimal value of the main global score of the metric.
        min_value: Optional[float] = None
        # The theorical maximal value of the main global score of the metric.
        max_value: Optional[float] = None

        # Public methods
        def compute(self) -> Any:
            return None

        def forward(self, *args: Any, **kwargs: Any) -> Any:
            self.update(*args, **kwargs)
            output = self.compute()
            self.reset()
            return output

        def reset(self) -> None:
            pass

        def update(self, *args, **kwargs) -> None:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import time

from functools import partial
from typing import Any, Callable, Iterable, Union

import torch

from torch import Tensor

from aac_metrics.functional.bleu import bleu
from aac_metrics.functional.cider_d import cider_d
from aac_metrics.functional.fense import fense
from aac_metrics.functional.fluerr import fluerr
from aac_metrics.functional.meteor import meteor
from aac_metrics.functional.rouge_l import rouge_l
from aac_metrics.functional.sbert_sim import sbert_sim
from aac_metrics.functional.spice import spice
from aac_metrics.functional.spider import spider
from aac_metrics.functional.spider_fl import spider_fl
from aac_metrics.utils.tokenization import preprocess_mono_sents, preprocess_mult_sents


pylog = logging.getLogger(__name__)


METRICS_SETS: dict[str, tuple[str, ...]] = {
    # Legacy metrics for AAC
    "default": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "spider",  # includes cider_d, spice
    ),
    # DCASE challenge task6a metrics for 2020, 2021 and 2022
    "dcase2020": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "spider",  # includes cider_d, spice
    ),
    # DCASE challenge task6a metrics for 2023
    "dcase2023": (
        "meteor",
        "spider_fl",  # includes cider_d, spice, spider, fluerr
    ),
    # All metrics
    "all": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "fense",  # includes sbert, fluerr
        "spider_fl",  # includes cider_d, spice, spider, fluerr
    ),
}


def evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: bool = True,
    metrics: Union[
        str, Iterable[str], Iterable[Callable[[list, list], tuple]]
    ] = "default",
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    device: Union[str, torch.device, None] = "auto",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with custom metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics.defaults to True.
    :param metrics: The name of the metric list or the explicit list of metrics to compute. defaults to "default".
    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param java_path: The path to the java executable. defaults to "java".
    :param tmp_path: Temporary directory path. defaults to "/tmp".
    :param device: The PyTorch device used to run FENSE and SPIDErFL models.
        If None, it will try to detect use cuda if available. defaults to "auto".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores.
    """
    metrics = _instantiate_metrics_functions(
        metrics, cache_path, java_path, tmp_path, device, verbose
    )

    if preprocess:
        candidates = preprocess_mono_sents(
            candidates,
            cache_path,
            java_path,
            tmp_path,
            verbose=verbose,
        )
        mult_references = preprocess_mult_sents(
            mult_references,
            cache_path,
            java_path,
            tmp_path,
            verbose=verbose,
        )

    outs_corpus = {}
    outs_sents = {}

    for i, metric in enumerate(metrics):
        if hasattr(metric, "__qualname__"):
            name = metric.__qualname__
        else:
            name = metric.__class__.__qualname__

        if verbose >= 1:
            pylog.info(f"[{i+1:2d}/{len(metrics):2d}] Computing {name} metric...")

        start = time.perf_counter()
        outs_corpus_i, outs_sents_i = metric(candidates, mult_references)
        end = time.perf_counter()

        if verbose >= 1:
            pylog.info(
                f"[{i+1:2d}/{len(metrics):2d}] Metric {name} computed in {end - start:.2f}s."
            )

        if __debug__:
            corpus_overlap = tuple(
                set(outs_corpus_i.keys()).intersection(outs_corpus.keys())
            )
            sents_overlap = tuple(
                set(outs_sents_i.keys()).intersection(outs_sents.keys())
            )
            if len(corpus_overlap) > 0 or len(sents_overlap) > 0:
                pylog.warning(
                    f"Found overlapping metric outputs names. (found {corpus_overlap=} and {sents_overlap=})"
                )

        outs_corpus |= outs_corpus_i
        outs_sents |= outs_sents_i

    return outs_corpus, outs_sents


def dcase2023_evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    device: Union[str, torch.device, None] = "auto",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with all Audio Captioning metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics.
        defaults to True.
    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param java_path: The path to the java executable. defaults to "java".
    :param tmp_path: Temporary directory path. defaults to "/tmp".
    :param device: The PyTorch device used to run FENSE and SPIDErFL models.
        If None, it will try to detect use cuda if available. defaults to "auto".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores.
    """
    return evaluate(
        candidates,
        mult_references,
        preprocess,
        "dcase2023",
        cache_path,
        java_path,
        tmp_path,
        device,
        verbose,
    )


def _instantiate_metrics_functions(
    metrics: Union[str, Iterable[str], Iterable[Callable[[list, list], tuple]]] = "all",
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    device: Union[str, torch.device, None] = "auto",
    verbose: int = 0,
) -> list[Callable]:
    if isinstance(metrics, str) and metrics in METRICS_SETS:
        metrics = METRICS_SETS[metrics]

    if isinstance(metrics, str):
        metrics = [metrics]
    else:
        metrics = list(metrics)  # type: ignore

    if not all(isinstance(metric, (str, Callable)) for metric in metrics):
        raise TypeError(
            "Invalid argument type for metrics. (expected str, Iterable[str] or Iterable[Metric])"
        )

    metric_factory = _get_metric_factory_functions(
        True,
        cache_path,
        java_path,
        tmp_path,
        device,
        verbose,
    )

    metrics_inst: list[Callable] = []
    for metric in metrics:
        if isinstance(metric, str):
            metric = metric_factory[metric]
        metrics_inst.append(metric)
    return metrics_inst


def _get_metric_factory_functions(
    return_all_scores: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    device: Union[str, torch.device, None] = "auto",
    verbose: int = 0,
) -> dict[str, Callable[[list[str], list[list[str]]], Any]]:
    return {
        "bleu_1": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=1,
        ),
        "bleu_2": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=2,
        ),
        "bleu_3": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=3,
        ),
        "bleu_4": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=4,
        ),
        "meteor": partial(
            meteor,
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            verbose=verbose,
        ),
        "rouge_l": partial(
            rouge_l,
            return_all_scores=return_all_scores,
        ),
        "cider_d": partial(
            cider_d,
            return_all_scores=return_all_scores,
        ),
        "spice": partial(
            spice,
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            verbose=verbose,
        ),
        "spider": partial(
            spider,
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            verbose=verbose,
        ),
        "sbert": partial(
            sbert_sim,
            return_all_scores=return_all_scores,
            device=device,
            verbose=verbose,
        ),
        "fluerr": partial(
            fluerr,
            return_all_scores=return_all_scores,
            device=device,
            verbose=verbose,
        ),
        "fense": partial(
            fense,
            return_all_scores=return_all_scores,
            device=device,
            verbose=verbose,
        ),
        "spider_fl": partial(
            spider_fl,
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            device=device,
            verbose=verbose,
        ),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
import subprocess
import sys

from argparse import ArgumentParser, Namespace
from subprocess import CalledProcessError

from torch.hub import download_url_to_file

from aac_metrics.classes.fense import FENSE
from aac_metrics.functional.meteor import FNAME_METEOR_JAR
from aac_metrics.functional.spice import FNAME_SPICE_JAR, DNAME_SPICE_CACHE
from aac_metrics.utils.tokenization import FNAME_STANFORD_CORENLP_3_4_1_JAR


pylog = logging.getLogger(__name__)


JAR_URLS = {
    "meteor": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/meteor/meteor-1.5.jar",
        "fname": "meteor-1.5.jar",
    },
    "meteor_data": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/meteor/data/paraphrase-en.gz",
        "fname": osp.join("data", "paraphrase-en.gz"),
    },
    "spice": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/spice/spice-1.0.jar",
        "fname": "spice-1.0.jar",
    },
    "stanford_nlp": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/tokenizer/stanford-corenlp-3.4.1.jar",
        "fname": "stanford-corenlp-3.4.1.jar",
    },
}
_TRUE_VALUES = ("true", "1", "t")
_FALSE_VALUES = ("false", "0", "f")


def download(
    cache_path: str = "$HOME/.cache",
    tmp_path: str = "/tmp",
    ptb_tokenizer: bool = True,
    meteor: bool = True,
    spice: bool = True,
    fense: bool = True,
    verbose: int = 0,
) -> None:
    """Download the code needed for SPICE, METEOR and PTB Tokenizer.

    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param tmp_path: The path to a temporary directory. defaults to "/tmp".
    :param ptb_tokenizer: If True, downloads the PTBTokenizer code in cache directory. defaults to True.
    :param meteor: If True, downloads the METEOR code in cache directory. defaults to True.
    :param spice: If True, downloads the SPICE code in cache directory. defaults to True.
    :param fense: If True, downloads the FENSE models. defaults to True.
    :param verbose: The verbose level. defaults to 0.
    """
    cache_path = osp.expandvars(cache_path)
    tmp_path = osp.expandvars(tmp_path)

    os.makedirs(cache_path, exist_ok=True)
    os.makedirs(tmp_path, exist_ok=True)

    if ptb_tokenizer:
        # Download JAR file for tokenization
        stanford_nlp_dpath = osp.join(
            cache_path, osp.dirname(FNAME_STANFORD_CORENLP_3_4_1_JAR)
        )
        os.makedirs(stanford_nlp_dpath, exist_ok=True)

        name = "stanford_nlp"
        info = JAR_URLS[name]
        url = info["url"]
        fname = info["fname"]
        fpath = osp.join(stanford_nlp_dpath, fname)
        if not osp.isfile(fpath):
            if verbose >= 1:
                pylog.info(
                    f"Downloading jar source for '{name}' in directory {stanford_nlp_dpath}."
                )
            download_url_to_file(url, fpath, progress=verbose >= 1)
        else:
            if verbose >= 1:
                pylog.info(f"Stanford model file '{name}' is already downloaded.")

    if meteor:
        # Download JAR files for METEOR metric
        meteor_dpath = osp.join(cache_path, osp.dirname(FNAME_METEOR_JAR))
        os.makedirs(meteor_dpath, exist_ok=True)

        for name in ("meteor", "meteor_data"):
            info = JAR_URLS[name]
            url = info["url"]
            fname = info["fname"]
            subdir = osp.dirname(fname)
            fpath = osp.join(meteor_dpath, fname)

            if not osp.isfile(fpath):
                if verbose >= 1:
                    pylog.info(
                        f"Downloading jar source for '{name}' in directory {meteor_dpath}."
                    )
                if subdir not in ("", "."):
                    os.makedirs(osp.join(meteor_dpath, subdir), exist_ok=True)

                download_url_to_file(
                    url,
                    fpath,
                    progress=verbose >= 1,
                )
            else:
                if verbose >= 1:
                    pylog.info(f"Meteor file '{name}' is already downloaded.")

    if spice:
        # Download JAR files for SPICE metric
        spice_jar_dpath = osp.join(cache_path, osp.dirname(FNAME_SPICE_JAR))
        spice_cache_path = osp.join(cache_path, DNAME_SPICE_CACHE)

        os.makedirs(spice_jar_dpath, exist_ok=True)
        os.makedirs(spice_cache_path, exist_ok=True)

        script_path = osp.join(osp.dirname(__file__), "..", "..", "install_spice.sh")
        if not osp.isfile(script_path):
            raise FileNotFoundError(
                f"Cannot find script '{osp.basename(script_path)}'."
            )

        if verbose >= 1:
            pylog.info(
                f"Downloading JAR sources for SPICE metric into '{spice_jar_dpath}'..."
            )

        command = ["bash", script_path, spice_jar_dpath]
        try:
            subprocess.check_call(
                command,
                stdout=None if verbose >= 2 else subprocess.DEVNULL,
                stderr=None if verbose >= 2 else subprocess.DEVNULL,
            )
        except (CalledProcessError, PermissionError) as err:
            pylog.error(err)

    if fense:
        # Download models files for FENSE metric
        if verbose >= 1:
            pylog.info("Downloading sBert and Bert error detector for FENSE metric...")
        _ = FENSE(device="cpu")


def _get_main_download_args() -> Namespace:
    parser = ArgumentParser(
        description="Download models and external code to evaluate captions."
    )

    parser.add_argument(
        "--cache_path",
        type=str,
        default="$HOME/.cache",
        help="Cache directory path.",
    )
    parser.add_argument(
        "--tmp_path",
        type=str,
        default="/tmp",
        help="Temporary directory path.",
    )
    parser.add_argument(
        "--ptb_tokenizer",
        type=_str_to_bool,
        default=True,
        help="Download PTBTokenizer Java source code.",
    )
    parser.add_argument(
        "--meteor",
        type=_str_to_bool,
        default=True,
        help="Download METEOR Java source code.",
    )
    parser.add_argument(
        "--spice",
        type=_str_to_bool,
        default=True,
        help="Download SPICE Java source code.",
    )
    parser.add_argument(
        "--fense",
        type=_str_to_bool,
        default=True,
        help="Download FENSE models.",
    )
    parser.add_argument("--verbose", type=int, default=1, help="Verbose level.")

    args = parser.parse_args()
    return args


def _main_download() -> None:
    format_ = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format_))
    pkg_logger = logging.getLogger("aac_metrics")
    pkg_logger.addHandler(handler)

    args = _get_main_download_args()

    level = logging.INFO if args.verbose <= 1 else logging.DEBUG
    pkg_logger.setLevel(level)

    download(
        args.cache_path,
        args.tmp_path,
        args.ptb_tokenizer,
        args.meteor,
        args.spice,
        args.fense,
        args.verbose,
    )


def _str_to_bool(s: str) -> bool:
    s = str(s).strip().lower()
    if s in _TRUE_VALUES:
        return True
    elif s in _FALSE_VALUES:
        return False
    else:
        raise ValueError(
            f"Invalid argument {s=}. (expected one of {_TRUE_VALUES + _FALSE_VALUES})"
        )


if __name__ == "__main__":
    _main_download()
